package nu.zoom.tapir;

import nu.zoom.tapir.parser.*;

import java.io.IOException;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.concurrent.Callable;

import picocli.CommandLine;
import picocli.CommandLine.Command;
import picocli.CommandLine.Option;
import picocli.CommandLine.Parameters;

@Command(name = "tapirgen", mixinStandardHelpOptions = true, description = "Generate source code from a tapir endpoint specification file.")
public class Generator implements Callable<Integer> {
	@Parameters(index = "0", description = "The source tapir file.")
	private Path file;

	@Option(names = {"-t", "--template"}, description = "The template directory. Default is ~/tapir-templates")
	private Path templateDir = Paths.get(System.getProperty("user.dir"), "tapir-templates");

	@Option(names = {"-o", "--output"}, description = "The directory to write the gerenated code to. Default is ~/tapir-output")
	private Path outputDir = Paths.get(System.getProperty("user.dir"), "tapir-output");

	@Option(names = {"-v", "--verbose"}, description = "Print verbose debug messages.")
	private Boolean verbose = false;

	public static void main(String[] args) throws ParseException {
		int exitCode = new CommandLine(new Generator()).execute(args);
		System.exit(exitCode);
	}

	@Override
	public Integer call() {
		try {
			validateTemplateDirectory();
			validateInputFile();
			validateOutputDirectory();
			var rootNode = new TapirParser(Files.newBufferedReader(this.file)).specification();
			if (this.verbose) {
				System.out.println("====== Parse Tree ======");
				rootNode.dump("");
			}
			NodeTransformer transformer = new NodeTransformer();
			transformer.transform(rootNode);
			if (transformer.getEndpoints().isEmpty()) {
				System.err.println("No tapir endpoints found.");
				return 2;
			}
			if (this.verbose) {
				System.out.println("\n====== AST ======");
				System.out.println("\n====== Types ======");
				transformer.getDataTypes().forEach(type -> {
					System.out.println(type);
				});
				System.out.println("\n====== Endpoints ======");
				transformer.getEndpoints().forEach(endpoint -> {
					System.out.println(endpoint);
				});
			}
			TargetGenerator targetGenerator = new TargetGenerator(
				this.verbose,
				this.outputDir,
				this.templateDir,
				transformer.getEndpoints(),
				transformer.getDataTypes()
			);
			targetGenerator.generate();
			return 0;
		} catch (Exception e) {
			System.err.println(e.getMessage());
			return 1;
		}
	}

	private void validateOutputDirectory() throws IOException {
		if (Files.notExists(this.outputDir)) {
			Files.createDirectories(this.outputDir);
		}
		if (!Files.isDirectory(this.outputDir)) {
			throw new IllegalArgumentException("Output directory: '" + this.outputDir + " 'is not a directory.");
		}
	}

	private void validateTemplateDirectory() throws IOException {
		if (!Files.isDirectory(this.templateDir)) {
			throw new IllegalArgumentException("Template directory '" + this.templateDir + "' does not exist.");
		}
		Path endpointTemplate = this.templateDir.resolve(TargetGenerator.ENDPOINTS_TEMPLATE_NAME);
		if (Files.notExists(endpointTemplate)) {
			throw new IllegalArgumentException("Can not find: '" + endpointTemplate + "'.");
		}
	}

	private void validateInputFile() throws IOException {
		if (Files.notExists(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' does not exist.");
		}
		if (!Files.isReadable(this.file) || !Files.isRegularFile(this.file)) {
			throw new IllegalArgumentException("Input file '" + this.file + "' is not a readable file.");
		}
	}
}
