package nu.zoom.tapir;

import nu.zoom.tapir.parser.Node;
import nu.zoom.tapir.parser.ParseException;
import nu.zoom.tapir.parser.SimpleNode;
import nu.zoom.tapir.parser.TapirParserTreeConstants;

import java.util.ArrayList;
import java.util.List;

public class NodeTransformer {
	private final List<EndpointNode> endpoints = new ArrayList<>();
	private final List<DataTypeNode> dataTypes = new ArrayList<>();

	public List<EndpointNode> getEndpoints() {
		return endpoints;
	}

	public List<DataTypeNode> getDataTypes() {
		return dataTypes;
	}

	public void transform(SimpleNode rootNode) throws ParseException {
		assertSimpleNodeType(rootNode, TapirParserTreeConstants.JJTSPECIFICATION);
		int numChildren = rootNode.jjtGetNumChildren();
		if (numChildren == 2) {
			this.dataTypes.addAll(
				handleDataTypes(
					assertSimpleNodeType(
						rootNode.jjtGetChild(0),
						TapirParserTreeConstants.JJTDATATYPES
					)
				)
			);
			this.endpoints.addAll(
				handleEndpoints(
					assertSimpleNodeType(
						rootNode.jjtGetChild(1),
						TapirParserTreeConstants.JJTENDPOINTS
					)
				)
			);
		} else if (numChildren == 1) {
			this.endpoints.addAll(
				handleEndpoints(
					assertSimpleNodeType(
						rootNode.jjtGetChild(1),
						TapirParserTreeConstants.JJTENDPOINTS
					)
				)
			);
		} else {
			throw new ParseException("Expected specification to have 1 or 2 children but had " + numChildren);
		}
	}

	private static List<EndpointNode> handleEndpoints(SimpleNode endpoints) throws ParseException {
		ArrayList<EndpointNode> endpointNodes = new ArrayList<>();
		for (int i = 0; i < endpoints.jjtGetNumChildren(); i++) {
			endpointNodes.add(
				handleEndpoint(
					assertSimpleNodeType(
						endpoints.jjtGetChild(i),
						TapirParserTreeConstants.JJTENDPOINT
					)
				)
			);
		}
		return endpointNodes;
	}

	private static List<DataTypeNode> handleDataTypes(SimpleNode dataTypesDeclaration) throws ParseException {
		List<DataTypeNode> dataTypes = new ArrayList<>();
		for (int i = 0; i < dataTypesDeclaration.jjtGetNumChildren(); i++) {
			dataTypes.add(
				handleCompoundDataType(
					assertSimpleNodeType(
						dataTypesDeclaration.jjtGetChild(i),
						TapirParserTreeConstants.JJTCOMPOUNDDATATYPE
					)
				)
			);
		}
		return dataTypes;
	}

	private static DataTypeNode handleCompoundDataType(SimpleNode dataTypeNode) throws ParseException {
		String typename = getStringValue(
			assertSimpleNodeType(
				dataTypeNode.jjtGetChild(0),
				TapirParserTreeConstants.JJTCOMPUNDDATATYPENAME
			)
		);
		List<FieldNode> fields = handleFields(
			assertSimpleNodeType(
				dataTypeNode.jjtGetChild(1),
				TapirParserTreeConstants.JJTDATATYPEFIELDS
			)
		);
		return new DataTypeNode(typename, fields);
	}

	private static EndpointNode handleEndpoint(SimpleNode node) throws ParseException {
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 2) {
			throw new ParseException("Endpoint node should have two children but had: " + numChildren);
		}
		SimpleNode pathsParseNode =
			assertSimpleNodeType(
				node.jjtGetChild(0),
				TapirParserTreeConstants.JJTPATH
			);
		PathsNode pathsNode = handlePaths(pathsParseNode);
		SimpleNode handlerParseNode =
			assertSimpleNodeType(
				node.jjtGetChild(1),
				TapirParserTreeConstants.JJTCOMPOUNDDATATYPE
			);
		HandlerNode handlerNode = handleHandler(handlerParseNode);
		return new EndpointNode(pathsNode, handlerNode);
	}

	private static HandlerNode handleHandler(SimpleNode handlerSpec) throws ParseException {
		int numChildren = handlerSpec.jjtGetNumChildren();
		if (numChildren != 2) {
			throw new ParseException("HandlerSpec-node should have two children but had: " + numChildren);
		}
		String handlerName = getStringValue(
			assertSimpleNodeType(
				handlerSpec.jjtGetChild(0),
				TapirParserTreeConstants.JJTCOMPUNDDATATYPENAME
			)
		);
		SimpleNode payloadFieldsParseNode =
			assertSimpleNodeType(
				handlerSpec.jjtGetChild(1),
				TapirParserTreeConstants.JJTDATATYPEFIELDS
			);
		List<FieldNode> fields = handleFields(payloadFieldsParseNode);
		return new HandlerNode(handlerName, fields);
	}

	private static List<FieldNode> handleFields(SimpleNode compoundDatatTypeFields) throws ParseException {
		ArrayList<FieldNode> fields = new ArrayList<>();
		for (int i = 0; i < compoundDatatTypeFields.jjtGetNumChildren(); i++) {
			SimpleNode payloadFieldParseNode =
				assertSimpleNodeType(
					compoundDatatTypeFields.jjtGetChild(i),
					TapirParserTreeConstants.JJTDATATYPEFIELD
				);
			int numFieldNodes = payloadFieldParseNode.jjtGetNumChildren();
			if (numFieldNodes != 2) {
				throw new ParseException("Field-node should have two children but had: " + numFieldNodes);
			}
			String fieldName = getStringValue(
				assertSimpleNodeType(
					payloadFieldParseNode.jjtGetChild(0),
					TapirParserTreeConstants.JJTDATATYPEFIELDNAME
				)
			);
			String fieldType = getStringValue(
				assertSimpleNodeType(
					payloadFieldParseNode.jjtGetChild(1),
					TapirParserTreeConstants.JJTDATATYPEFIELDTYPE
				)
			);
			fields.add(new FieldNode(fieldName, fieldType));
		}
		return fields;
	}

	private static PathsNode handlePaths(SimpleNode pathsParseNode) throws ParseException {
		int numPathSegments = pathsParseNode.jjtGetNumChildren();
		ArrayList<String> segments = new ArrayList<>();
		for (int i = 0; i < numPathSegments; i++) {
			SimpleNode segmentParseNode = assertSimpleNodeType(pathsParseNode.jjtGetChild(i), TapirParserTreeConstants.JJTPATHSEGMENT);
			segments.add(getStringValue(segmentParseNode));
		}
		return new PathsNode(segments);
	}

	private static SimpleNode assertSimpleNodeType(Node node, int nodeType) throws ParseException {
		SimpleNode result = assertSimpleNode(node);
		if (result.getId() != nodeType) {
			throw new ParseException(
				"Expected Node of type " +
					TapirParserTreeConstants.jjtNodeName[nodeType]
					+ " but was " + TapirParserTreeConstants.jjtNodeName[result.getId()]
			);
		}
		return result;
	}

	private static SimpleNode assertSimpleNode(Node node) throws ParseException {
		if (node instanceof SimpleNode) {
			return (SimpleNode) node;
		}
		throw new ParseException("Expected Node to be a " + SimpleNode.class + " but found " + node.getClass());
	}

	private static String getStringValue(SimpleNode node) throws ParseException {
		Object value = node.jjtGetValue();
		if (value instanceof String) {
			return (String) value;
		}
		throw new ParseException("Expected Node " + node + " to have a String value but was " + value.getClass());
	}
}
