package nu.zoom.tapir;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;

public class TargetGenerator {
	private final Path outputPath;
	private final Path templatePath;
	public static String ENDPOINTS_TEMPLATE_NAME = "endpoints.ftl";
	private final List<EndpointNode> endpoints;
	private final List<DataTypeNode> dataTypes;
	private final boolean verbose ;

	public static class TargetGeneratorException extends Exception {
		public TargetGeneratorException(Exception cause) {
			super(cause);
		}
	}

	public TargetGenerator(
		boolean verbose,
		Path outputPath,
		Path templatePath,
		List<EndpointNode> endpoints,
		List<DataTypeNode> dataTypes
	) {
		this.verbose = verbose;
		this.outputPath = Objects.requireNonNull(
			outputPath,
			"Output path is required"
		);
		this.templatePath = Objects.requireNonNull(
			templatePath,
			"Template path is required"
		);
		this.endpoints = Objects.requireNonNull(endpoints);
		this.dataTypes = Objects.requireNonNull(dataTypes);
	}

	public void generate() throws TargetGeneratorException {
		try {
			Configuration cfg = new Configuration(Configuration.VERSION_2_3_34);
			cfg.setDirectoryForTemplateLoading(this.templatePath.toFile());
			cfg.setDefaultEncoding("UTF-8");
			cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
			cfg.setLogTemplateExceptions(false);
			cfg.setWrapUncheckedExceptions(true);
			cfg.setFallbackOnNullLoopVariable(false);
			Template temp = cfg.getTemplate(ENDPOINTS_TEMPLATE_NAME);
			List<Path> templates = Files
				.list(this.templatePath)
				.filter(Files::isRegularFile)
				.filter(f -> f.getFileName().toString().endsWith(".ftl"))
				.toList() ;
			for (Path template : templates) {
				try (var outputFile = Files.newBufferedWriter(
					outputName(template),
					StandardOpenOption.CREATE,
					StandardOpenOption.TRUNCATE_EXISTING
				)) {
					if (this.verbose) {
						System.out.println("Processing " + template);
					}
					HashMap<String, Object> templateData = new HashMap<>();
					templateData.put("endpoints", endpoints);
					templateData.put("datatypes", dataTypes);
					temp.process(templateData, outputFile);
				}
			}
		} catch (TemplateException | IOException ex) {
			throw new TargetGeneratorException(ex);
		}
	}

	private Path outputName(Path templatePath) {
		String name = templatePath.getFileName().toString().replace(".ftl", ".scala") ;
		return this.outputPath.resolve(name) ;
	}
}
